import java.net.URL;
import java.sql.*;
import java.io.*;
import java.util.*;

import javax.servlet.*;
import javax.servlet.http.*;

/**
 * A search servlet demo
 *
 * It supports AND and OR query, but it does not return a detailed summary.
 *
 * This demo is suitable for the limited hosts or recursive depth since it maintains
 * its list of URLs and keywords in memory.
 *
 * This search engine demo is not suitable for indexing
 * a large number of URLs since it holds the information
 * of URLs and keywords in memory. If you wish to do so,
 * you should seek the relative information of the specific
 * keyword (or URL) in the index database through JDBC 
 * interface, not in memory through hash tables. 
 *
 * I've included the retrieval reports for http://www.mokabyte.it
 * and file:///d:/article/firewallutil/ with a recursive depth of 3.
 * Some query samples are also provided in HTML format.
 */

public class Search extends HttpServlet
{
	private static final String servletName="Search";

	private static final String title="A Search Servlet Demo Version 1.0";
	public String getServletInfo()
	{return title+"\n(c) Zhao Yonghong\nJune 2000";}

 	/*
	 * The hashtable containing all visited links.
	 * It's possible to load all visited links into memory
	 * since we're using it for the limited hosts or recursive depth.
	 */
	private Hashtable urlsTable=null;
 	/*
	 * The hashtable containing all indexed keywords.
	 * It's possible to load all indexed keywords into memory
	 * since we're constructed a full-text search
	 * engine for the limited hosts or recursive depth.
	 */
	private Hashtable keywordTable;
	private String keywordTableName;

	private HttpServletRequest request=null;
	private HttpServletResponse response=null;

	private ByteArrayOutputStream bytes=null;
	private	PrintStream	outstream=null;

	//The JDBC Connection for the index database of URLs and keywords
	private Connection connection=null;
	
	/**
	 * Initialize JDBC driver and url parameters
	 *
	 * It need a JDBC 2.0 driver which supports the scrollable
	 * updatable resultset. Some serializable objects will be 
	 * holden in some OBJECT or VARBINARY columns.
	 *
	 * The default demo driver is <a href=http://members.tripod.com/~ZYH/jdbc.html>DBF JDBC driver</a>
	 *
	 * Search.initparams=driver=zyh.sql.dbf.DBFDriver,jdbcurl=jdbc:DBF:/.
	 *
	 */
	public void init(ServletConfig servletConfig)
	{
		try
		{
			super.init(servletConfig);
			bytes=new ByteArrayOutputStream (4096);
			outstream=new PrintStream (bytes);

			String driverName = getInitParameter("driver");
			if (driverName == null)
				driverName="zyh.sql.dbf.DBFDriver";

			String jdbcurl=getInitParameter("jdbcurl");
			if (jdbcurl == null)
				jdbcurl="jdbc:DBF:/c:/search/";

			String urlsTableName=getInitParameter("urstable");
			if (urlsTableName == null)
				urlsTableName="urls";

			keywordTableName=getInitParameter("keywordtable");
			if (keywordTableName == null)
				keywordTableName="keywords";


			Class.forName(driverName).newInstance();
			connection=DriverManager.getConnection(jdbcurl,"","");

			Statement stmt = connection.createStatement();
			stmt.setFetchSize(100);

			ResultSet rs=stmt.executeQuery("select max(urlID) from "+urlsTableName);
			int urlsum=1;
			if(rs.next())
				urlsum=rs.getInt(1);
			rs.close();

			urlsTable=new Hashtable(urlsum);

			rs=stmt.executeQuery("select url,urlID from "+urlsTableName);
			while (rs.next())
			{
				urlsTable.put(new Integer(rs.getInt(2)),rs.getString(1));
			}
			rs.close();

			rs=stmt.executeQuery("select max(wordID) from "+keywordTableName);
			int wordsum=1;
			if(rs.next())
				wordsum=rs.getInt(1);
			rs.close();

			keywordTable=new Hashtable(wordsum);

			rs=stmt.executeQuery("select keyword,wordID from "+keywordTableName);
			while (rs.next())
			{
				keywordTable.put(rs.getString(1),new Integer(rs.getInt(2)));
			}
			rs.close();
		}
		catch (Exception e)
		{
			e.printStackTrace();
			System.out.println(e);
		}
	}
	private String getURL(int urlID)
	{
		return (String)urlsTable.get(new Integer(urlID));
	}
	private int getWordID(String word)
	{
		Object obj=keywordTable.get(word.toLowerCase());//Use lowercase since the keywords in database are lowercase
		if(obj==null)
			return -1;
		else
			return ((Integer)obj).intValue();
	}

	public void destroy()
	{
		try
		{
			if (connection!=null)
			{
				connection.close();
				connection=null;
			}
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
	}

	public synchronized void service(HttpServletRequest requestp, HttpServletResponse response)throws ServletException, IOException
	{
		Statement statement=null;
		try
		{
			this.request=requestp;			
			this.response=response;

			setContentType(); //Avoid the browser reset the connection after a long time wait.

			String keywordp=request.getParameter("keywords");
			String operator=request.getParameter("operator");
			if(keywordp==null || operator==null)
			{
				outputHTML(getSearchForm());
				return;
			}

			StringTokenizer st = new StringTokenizer(keywordp);
			ArrayList al=new ArrayList(5);
			while (st.hasMoreTokens())
				al.add(st.nextToken());
      		if(al.size()==0)
			{
				outputHTML(getSearchForm());
				return;
			}

			String[] keywords=new String[al.size()];
			int[] wordIDs=new int[keywords.length];

			String sql=null;
			String[] sqls=null;

			keywords[0]=(String)al.get(0);
			wordIDs[0]=getWordID(keywords[0]);

			StringBuffer sqlsb=new StringBuffer("SELECT urls FROM "+keywordTableName+" WHERE "
				+"wordID="+wordIDs[0]);

			if(operator.equalsIgnoreCase("OR"))//OR operator
			{
				for(int i=1;i<keywords.length;i++)
				{
					keywords[i]=(String)al.get(i);
					wordIDs[i]=getWordID(keywords[i]);
					sqlsb.append(" "+operator+" ");
					sqlsb.append("wordID="+wordIDs[i]);
				}
				sql=sqlsb.toString();
			}
			else//AND operator
			{
				sql=sqlsb.toString();
				sqls=new String[keywords.length-1];
				for(int i=1;i<keywords.length;i++)
				{
					keywords[i]=(String)al.get(i);
					wordIDs[i]=getWordID(keywords[i]);
					sqls[i-1]="SELECT urls FROM "+keywordTableName+" WHERE "
						+"wordID="+wordIDs[i];
				}
			}	
			statement=connection.createStatement();
			statement.setMaxRows(100);//Set the maximum number of rows that a ResultSet can contain.
			ResultSet resultSet=statement.executeQuery(sql);
			if (resultSet!=null)
			{
				int[] urlIDs;
				if(sqls==null)//OR operator
					urlIDs=zyh.robot.ContentSorter.sortURLs(1,resultSet);
				else//AND operator
				{
					urlIDs=zyh.robot.ContentSorter.sortURLs(1,resultSet,sqls,statement);
				}
				resultSet=null;

				outputHTML(getSearchForm(getQueryHTML(keywords,operator,urlIDs)));

				statement.close();
				statement=null;
			}
		}
		catch (SQLException e)
		{
			if (statement!=null)
			{
				try
				{
					statement.close();
				}
				catch(SQLException se){}

				statement=null;
			}
			outputHTML(getSearchForm(getSQLExceptionHTML(e)));
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			ArrayList al=new ArrayList(20);
			al.add(e+"");
			outputHTML(getSearchForm(al));
		}
	}


	private void setContentType()
	{
		response.setContentType("TEXT/HTML");
		outstream.print( "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN//\">");
	}

	private ArrayList getQueryHTML(String[] keywords,
		String operator,
		int[] urlIDs)
	{
		ArrayList words=new ArrayList(160);

		words.add("<h2 align=center><font color=#993366>Search Results</font></h2>");
		words.add("<HR WIDTH=100%>");	

		words.add("<font color=#800040><BR>Your query: </font>");
		for(int i=0;i<keywords.length-2;i++)
			words.add(keywords[i]+", ");
		if(keywords.length>1)
			words.add(keywords[keywords.length-2]
				+" "+operator+" ");
		words.add(keywords[keywords.length-1]
			+".<BR>");
		
		if(urlIDs.length==0)
			words.add("The search demo can not find any suitable URL.");
		else
		{
			words.add("The search demo has found "+urlIDs.length+" URL"+(urlIDs.length>1?"s.":".")+"<br>");
			String url;
			int number=urlIDs.length>25?25:urlIDs.length;
			for(int i=0; i <number; i++)
			{
				url=getURL(urlIDs[i]);
				words.add("<a href=\""+url+"\">"+url+"</a><br>");
			}
		}
		words.add("<HR WIDTH=100%>");

		return words;
	}

	private ArrayList getSQLExceptionHTML(SQLException e)
	{
		ArrayList words=new ArrayList(30);

		words.add("<h2 align=center><font color=#993366>A SQL Exception Occurred</h2>");
		
		while (e!=null)
		{
			words.add("<P>SQLState:  "+e.getSQLState()+"</P>");
			words.add("<P>Message:   "+e.getMessage()+"</P>");
			words.add("<P>ErrorCode: "+e.getErrorCode()+"</P>");
			e=e.getNextException();
		}
		words.add("<HR WIDTH=80%>");

		return words;
	}

	private static final String[] operators={"AND","OR"};
	private ArrayList getSearchForm()
	{
		return getSearchForm(new ArrayList(15));
	}
	private ArrayList getSearchForm(ArrayList words)
	{
		words.add("<h2 align=center><font color=#223344>" + title + "</h2></font>");
		words.add("<form action=http://"+request.getServerName()
			+":"+request.getServerPort()+"/servlet/"+servletName+" method=POST>");
		words.add("<blockquote><div align=center><center><p>Keywords: <textarea rows=2 name=keywords cols=49 tabindex=1></textarea></p>");
		words.add("</center></div></blockquote>");

		words.add("<blockquote><div align=center><center><p>Operator: <select name=operator size=1 tabindex=2>");
		words.add("<option selected value=" + operators[0] + ">" + operators[0] + "</option>");
		for (int i=1; i < operators.length; i++)
			words.add("<option value=" + operators[i] + ">" + operators[i] + "</option>");
		words.add("</select>&nbsp;&nbsp;</center></div></blockquote>");

		words.add("<blockquote><div align=center><center><p><input type=submit value=Submit Query tabindex=3><input type=reset value=\"Clear\" name=\"clear\" tabindex=4></p>");

		return words;
	}

	private void outputHTML(ArrayList words)
	{
		try
		{
			outstream.print("<HTML><HEAD><TITLE>"+title+"</TITLE></HEAD>");
			outstream.print("<BODY bgcolor=#C0C0C0 text=#FFFFFFF>");

			for(int i=0;i<words.size();i++)
				outstream.print((String)words.get(i));

			outstream.print("</BODY></HTML>");
			outstream.flush();
			bytes.writeTo(response.getOutputStream ());
	   		bytes.reset();
		}
		catch (IOException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
	}
}