/*
 *
 *  @author  Zhao Yonghong (zhaoyh@bigfoot.com)
 *
 *	Copyright 2000 by HXKM Inc., 9 Station Road, Xiangtan, 411100, P.R. China
 *	All rights reserved.
 * 
 *  You can distribute freely it without modification.
 *  You also can modify and distribute it but please keep our statements unchanged.
 *  For more licences conditions, please refer to
 *  <A HREF="http://www.fsf.org/copyleft/gpl.html">GNU General Public License (GPL)</A>
 *	or
 *  <A HREF="http://language.perl.com/misc/Artistic.html">Artistic License</A>
 *
 */

package zyh.net.http;

import java.net.*;
import java.io.*;
import java.util.*;

/**
 * HttpURLConnection class implements java.net.HttpURLConnection
 *
 * @see java.net.HttpURLConnection
 */
public class HttpURLConnection extends java.net.HttpURLConnection
{
	private long birthtime=0;

	private Socket clientSocket=null;
	private Hashtable headers=null;
	private ArrayList headerFields=null;
	private ArrayList headerFieldKeys=null;
	private Hashtable requests=null;

	private InetAddress proxyInetAddress=null;
	private int proxyPort=8080;

	//Please provide your email so that server maintainers can contact you in case of problems.
	private String mailbox=null;

	/**
	 * Constructor for HttpURLConnection object.
	 * @param url the destination url
	 */
	public HttpURLConnection(URL url)
		throws MalformedURLException
	{this(url,null,null,8080);}
	/**
	 * Constructor for HttpURLConnection object.
	 * @param url the destination url
	 * @param mailbox please provide your mailbox so that server maintainers can contact you in case of problems
	 */
	public HttpURLConnection(URL url,String mailbox)
		throws MalformedURLException
	{this(url,mailbox,null,8080);}
	/**
	 * Constructor for HttpURLConnection object.
	 * @param url the destination url
	 * @param mailbox please provide your mailbox so that server maintainers can contact you in case of problems
	 * @param proxyInetAddress the InetAddress of HTTP proxy
	 * @param proxyPort the port of HTTP proxy
	 */
	public HttpURLConnection(URL url,
		String mailbox,
		InetAddress proxyInetAddress,
		int proxyPort
		)throws MalformedURLException
	{
		super(url);

		this.url=url;
		this.mailbox=mailbox;
		this.proxyInetAddress=proxyInetAddress;
		this.proxyPort=proxyPort;
		if(this.proxyPort==-1)// proxyPort is not set.
			this.proxyPort=8080;
		this.birthtime=System.currentTimeMillis();

		// Do some sanity checking on the URL.
		if (!url.getProtocol().equals("http"))
		{
			throw new MalformedURLException("protocol must be \"http\", but you used "
				+url.getProtocol());
		}
		else if (url.getHost().length() == 0)
		{
			throw new MalformedURLException("no host name specified in "+url);
		}
	}

	/**
	 * Returns the value of this URLConnection's URL field.
	 */
	private URL url;
	public URL getURL()
	{
		return url;
	}
	/**
	 * Returns an input stream that reads from this open connection.
	 */
	public InputStream getInputStream() throws IOException
	{
		if(clientSocket!=null)
			return clientSocket.getInputStream();
		else
			throw new IllegalAccessError("No connected socket is available.");
	}

	/**
	 * Returns an output stream that writes to this connection.
	 */
	public OutputStream getOutputStream() throws IOException
	{
		if(clientSocket!=null)
			return clientSocket.getOutputStream();
		else
			throw new IllegalAccessError("No connected socket is available.");
	}

	/**
	 * Sets the logging/tracing Stream
	 *
	 * @param out the new logging/tracing PrintStream; to disable, set to null
	 */
	private PrintStream logStream=null;//System.out;
	public void setLog(PrintStream logStream)
	{this.logStream=logStream;}
	/**
	 * Gets the log stream.	
	 */
	public PrintStream getLog()
	{return logStream; }

	/**
	 * Prints a message to the current log stream.
	 *
	 * @param message a log or tracing message
	 */
	public void log(String msg)
	{
		if(logStream!=null)
			logStream.println(
				getURL()
				+"("
				+(new java.util.Date(System.currentTimeMillis())).toString()
				+(usingProxy()?" "+proxyInetAddress.getHostAddress()+":"+proxyPort:"")
				+")"
				+": "+msg);
	}

	/**
	 * Indicates if the connection is going through a proxy.
	 */
	public boolean usingProxy()
	{return proxyInetAddress!=null;}
	/**
	 * Close the connection to the server.
	 */
	public synchronized void disconnect()
	{
		if (clientSocket != null)
		{
			try
			{
				clientSocket.close();
				clientSocket=null;
				headers=null;
				headerFields=null;
				headerFieldKeys=null;
				requests=null;
			}
			catch(java.io.IOException e)
			{
				log(e.getMessage());
			}
		}
		if (connected)
		{
			connected = false;
		}
	}

	/**
	 * Opens a communications link to the resource referenced by this URL, if such a
	 * connection has not already been established.
	 */
	public synchronized void connect()throws IOException
	{
		connect(0);
	}
	private synchronized void connect(int redirectionCounter)throws IOException
	{
		int sleeptime=1000;
		for (int i=redirectionCounter;;i++)
		{
			try
			{
				doHTTPConnect();
				if(getFollowRedirects()
					&& (getResponseCode()== HTTP_MOVED_PERM || getResponseCode()==HTTP_MOVED_TEMP)
					&& redirectionCounter<=5 //Avoid infinite redirection loops
					)
				{
					String location=getHeaderField("Location");
					URL movedURL=new URL(url,location);
					HttpURLConnection connection=new HttpURLConnection(movedURL,mailbox,proxyInetAddress,proxyPort);
					connection.connect(redirectionCounter+1);
					clientSocket=connection.clientSocket;
					url=movedURL;
				}
				return;
			}
			catch (IOException e)
			{
				disconnect();
				if(i>=8 ||
					System.currentTimeMillis()-birthtime>180000)
				{
					log("Failed Connection Timeout - "+((System.currentTimeMillis()-birthtime)/1000)+" seconds.");
					throw e;
				}
				try
				{
					Thread.sleep(sleeptime);
					sleeptime+=600;
				}
				catch (InterruptedException ie){} 
			}
		}
	}

	private synchronized void doHTTPConnect()throws IOException
	{
		if(usingProxy())
			clientSocket = new Socket(proxyInetAddress, proxyPort);
		else
		{
			int port=getURL().getPort();
			if(port==-1)port=80;
			clientSocket = new Socket(InetAddress.getByName(getURL().getHost()),port);
		}

		clientSocket.setSoTimeout(180000);//Throw a java.io.InterruptedIOException if the timeout expires.

		outputRequest();

		initHeaders();

		connected = true;
	}

	private void outputRequest()throws IOException
	{
		PrintStream out= new PrintStream(clientSocket.getOutputStream());
		if(usingProxy())
			out.println(getRequestMethod()+" "
				+ getURL().toString()
				+ " HTTP/1.0");
		else
			out.println(getRequestMethod()+" "
				+ getURL().getFile()
				+ " HTTP/1.0");
		if(getRequestProperty("Referer")!= null)
			out.println("Referer: "+getRequestProperty("Referer"));

		if(usingProxy())
			out.println("Proxy-Connection: Keep-Alive");
		else
			out.println("Connection: Keep-Alive");

		out.println("User-Agent: Java Robot/1.01 (Linux; I)");
		//Please use your email so that server maintainers can contact you in case of problems.
		if(mailbox!=null)
			out.println("From: "+mailbox);

		if(usingProxy())
			out.println("Pragma: no-cache");
		int port=getURL().getPort();
		out.println("Host: "+getURL().getHost()
			+(port==-1?"":":"+port));

		out.println("Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, image/png, */*");
		out.println("Accept-Encoding: gzip");
		out.println("Accept-Language: en");
		out.println("Accept-Charset: iso-8859-1,*,utf-8");

		if(getRequestProperty("Cookie")!= null)
			out.println("Cookie: "+getRequestProperty("Cookie"));

		if(getRequestMethod().equals("POST"))
		{
			String content=getRequestProperty("Content");
			out.println("Content-type: application/x-www-form-urlencoded");
			out.println("Content-length: "+(content==null?0:content.length()));
			out.println();

			out.print(content);
		}
		else
			out.println();

		out.flush();
	}

	/*
		CR= <US-ASCII CR, carriage return (13)>
		LF= <US-ASCII LF, linefeed (10)>
		HTTP/1.1 defines the sequence CR LF
		as the end-of-line marker for all protocol
		elements except the entity-body (see appendix
		19.3 for tolerant applications). 
		The end-of-line marker within an entity-body 
		is defined by its associated media type, as described in section 3.7.
		This flexibility regarding line breaks applies only to
		text media in the entity-body; a bare CR or LF
		MUST NOT be substituted for CRLF within any of the HTTP control structures
		(such as header fields and multipart boundaries).
		CRLF		   = CR LF
	*/
	//Only use readLine function to read HTTP Header
	private static final String readLine(InputStream inputStream,
		byte bArray[])throws IOException
	{
		int readByteNumber;
		int offset=0;
		byte priorByte=-1;
		while(true)
		{
			readByteNumber=inputStream.read(bArray,offset,1);
			if(readByteNumber<0)
			{
				throw new IOException("Invalid HTTP header: Maybe socket time out.");
			}
			else if(readByteNumber>0)
			{
				if(bArray[offset]== '\n'
//					&& priorByte=='\r'//Seems some web servers only use LF
					)
					break;
				priorByte=bArray[offset];
				offset++;//offset+=readByteNumber;
			}
		}
		if(offset>=1 && bArray[offset]=='\n')
		{
			if(priorByte=='\r')
				return new String(bArray,0,offset-1);
			else
				return new String(bArray,0,offset);
		}
		else
			return new String(bArray,0,offset+1);
	}

	private void initHeaders()throws IOException
	{
		InputStream inputStream=  getInputStream();

		headers= new Hashtable(20);
		headerFields=new ArrayList(20);
		headerFieldKeys=new ArrayList(20);
		String line;
		String key,field;

		byte bArray[] = new byte[4096];

		while((line = readLine(inputStream,bArray)) != null
			&& line.length()>0)
		{
			int index = line.indexOf(": ");
			if (index > 0)
			{
				key=line.substring(0, index).toLowerCase();
				field=line.substring(index+2);
				headers.put(key,field);
			}
			else
			{
				key=null;
				field=line;
			}
			headerFields.add(field);
			headerFieldKeys.add(key);
//System.out.println(key+" "+field);
		}
//		if(line==null || line.length()!=0)
//			System.out.println(line+"Error");
	}
	/**
	 * Returns the name of the specified header field.
	 *
	 * @param   name   the name of a header field.
	 * @return  the value of the named header field, or <code>null</code>
	 *		  if there is no such field in the header.
	 */
	public String getHeaderField(String name)
	{
		return headers==null?null:(String)headers.get(name.toLowerCase());
	}

	/**
	 * Returns the key for the <code>n</code><sup>th</sup> header field.
	 *
	 * @param   n   an index.
	 * @return  the key for the <code>n</code><sup>th</sup> header field,
	 *		  or <code>null</code> if there are fewer than <code>n</code>
	 *		  fields.
	 */
	public String getHeaderFieldKey(int n)
	{
		return headers==null?null:(headerFieldKeys.size()>n?(String)headerFieldKeys.get(n):null);
	}

	/**
	 * Returns the value for the <code>n</code><sup>th</sup> header field. 
	 * It returns <code>null</code> if there are fewer than 
	 * <code>n</code> fields. 
	 * <p>
	 * This method can be used in conjunction with the 
	 * <code>getHeaderFieldKey</code> method to iterate through all 
	 * the headers in the message. 
	 *
	 * @param   n   an index.
	 * @return  the value of the <code>n</code><sup>th</sup> header field.
	 * @see	 java.net.URLConnection#getHeaderFieldKey(int)
	 */
	public String getHeaderField(int n)
	{
		return headers==null?null:(headerFields.size()>n?(String)headerFields.get(n):null);
	}

	/**
	 * Returns the value of the named general request property for this
	 * doHTTPConnect.
	 *
	 * @return  the value of the named general request property for this
	 *		   doHTTPConnect.
	 */
	public String getRequestProperty(String key)
	{
		if (connected)
			throw new IllegalAccessError("Already connected");
		if(requests==null)return null;
		else
			return (String)requests.get(key);
	}

	/**
	 * Sets the general request property. 
	 *
	 * @param   key	 the keyword by which the request is known
	 *				  (e.g., "<code>accept</code>").
	 * @param   value   the value associated with it.
	 */
	public void setRequestProperty(String key, String value)
	{
		if(connected)
			throw new IllegalAccessError("Already connected");
		if(key==null)return;
		if(requests==null)requests= new Hashtable(20);
		if(value==null)requests.remove(key);
		else
			requests.put(key,value);
	}

	/**
	 * Sets the default value of a general request property. When a 
	 * <code>URLConnection</code> is created, it is initialized with 
	 * these properties. 
	 *
	 * @param   key	 the keyword by which the request is known
	 *				  (e.g., "<code>accept</code>").
	 * @param   value   the value associated with the key.
	 */
	public static void setDefaultRequestProperty(String key, String value) {
	}

	/**
	 * Returns the value of the default request property. Default request 
	 * properties are set for every doHTTPConnect. 
	 *
	 * @return  the value of the default request property for the specified key.
	 * @see	 java.net.URLConnection#setDefaultRequestProperty(java.lang.String, java.lang.String)
	 */
	public static String getDefaultRequestProperty(String key) {
	return null;
	}


  /**
	* Returns the error stream if the doHTTPConnect failed
	* but the server sent useful data nonetheless. The
	* typical example is when an HTTP server responds
	* with a 404, which will cause a FileNotFoundException 
	* to be thrown in connect, but the server sent an HTML
	* help page with suggestions as to what to do.
	*
	* <p>This method will not cause a doHTTPConnect to be initiated.
	* If there the doHTTPConnect was not connected, or if the server
	* did not have an error while connecting or if the server did
	* have an error but there no error data was sent, this method
	* will return null. This is the default.
	*
	* @return an error stream if any, null if there have been
	* no errors, the doHTTPConnect is not connected or the server
	* sent no useful data.
	*/
	public InputStream getErrorStream(){
		return null;
	}

	/** A simple test application */
	public static void main(String args[]) throws Exception
	{
		try
		{
			String spec="http://127.0.0.1";
			if(args.length==1)
				spec=args[0];
			URL url=new URL(spec);
			zyh.net.http.HttpURLConnection connection=new zyh.net.http.HttpURLConnection(url,null);

			connection.setRequestMethod("GET");
	//		connection.setRequestMethod("POST");

			connection.setRequestProperty("Referer","http://abc");
//			connection.setRequestProperty("say","ddddd");
//			connection.setRequestProperty("submit","zzz");

			connection.connect();

			BufferedReader r = new BufferedReader(new InputStreamReader(connection.getInputStream()));
			char cbuf[] = new char[256];
			int nb;
			int nCount=0;
			while((nb = r.read(cbuf, 0, 255)) >= 0)
			{
				if(nb > 0)
				{
					System.out.print(new String(cbuf, 0, nb));
					nCount+=nb;
				}
			}
			r.close();

			System.out.println("nCount:"+nCount);

			connection.disconnect();
		}
		catch(Exception e)
		{
			e.printStackTrace();
			System.out.println(e.getMessage());
		}
	}
}