/*
 *
 *  @author  Zhao Yonghong (zhaoyh@bigfoot.com)
 *
 *	Copyright 2000 by HXKM Inc., 9 Station Road, Xiangtan, 411100, P.R. China
 *	All rights reserved.
 * 
 *  You can distribute freely it without modification.
 *  You also can modify and distribute it but please keep our statements unchanged.
 *  For more licences conditions, please refer to
 *  <A HREF="http://www.fsf.org/copyleft/gpl.html">GNU General Public License (GPL)</A>
 *	or
 *  <A HREF="http://language.perl.com/misc/Artistic.html">Artistic License</A>
 *
 */

package zyh.robot;

import java.io.*;
import java.util.*;

import java.net.URLConnection;
import java.net.URL;

import java.sql.SQLException;

/**
 * <P>The basic service for managing a set of content filters and processors.
 *
 * zyh.robot.ContentManager is with responsibility for registering manifold
 * content filters and content processors. The purpose of content filter is 
 * to parse raw data in some formats, such as text, HTML, PostScript or RTF,
 * and stores filtered information and extracted URLs in zyh.robot.Content, 
 * a data structure, which holds title, keywords, summary, text and extracted
 * URLs. Content processor decides on the index policy and does actual index 
 * process on the Content object, which is constructed by content filter.
 *
 * <P>When the method <code>getContent</code> is called,
 * the ContentManager will attempt to
 * locate a suitable content filter from those content filters
 * registered explicitly by the method <code>registerContentFilter</code>.
 *
 * <P>When the method <code>getWordIDs</code> is called,
 * the ContentManager will attempt to
 * locate a suitable content processor from those content processors
 * registered explicitly by the method <code>registerContentProcessors</code>.

 * @see Content
 * @see ContentFilter
 * @see ContentProcessor
 */
public class ContentManager
{
	/**
	 * Sets the logging/tracing Writer that is used by the ContentManager
	 * and all ContentFilter.
	 *
	 * @param out the new logging/tracing PrintStream; to disable, set to null
	 */
	private static java.io.PrintWriter logWriter = null;
	public static void setLogWriter(java.io.PrintWriter out)
	{
		logWriter = out;
	}

	/**
	 * Gets the log writer.	
	 */
	public static java.io.PrintWriter getLogWriter()
	{
		return logWriter;
	}

	/**
	 * Prints a message to the current ContentFilter log stream.
	 *
	 * @param message a log or tracing message
	 */
	public static synchronized void println(String message)
	{
		if (logWriter != null)
		{
			logWriter.println(message);
			logWriter.flush();
		}
	}

	/**
	 * Registers the content filter with the ContentManager.
	 *
	 * @param contentFilter the new content filter that is to be registered with the ContentManager
	 */
	private static java.util.Hashtable contentFilters = new java.util.Hashtable(10);
	public static synchronized void registerContentFilter(ContentFilter contentFilter)
	{	
		String[] acceptedContenttypes=contentFilter.getAcceptedContenttype();
		if(acceptedContenttypes==null)//It should do not occur.
			return;
		for(int i=0;i<acceptedContenttypes.length;i++)
			registerContentFilter(acceptedContenttypes[i],contentFilter);
	}
	/**
	 * Registers the content filter with the ContentManager.
	 *
	 * @param acceptedContenttype the accepted content type
	 * @param contentFilter the new content filter that is to be registered with the ContentManager
	 */
	public static synchronized void registerContentFilter(String acceptedContenttype,ContentFilter contentFilter)
	{	
		if(acceptedContenttype==null)return;
		ContentFilter oldContentFilter=(ContentFilter)contentFilters.put(acceptedContenttype.toLowerCase(),contentFilter);
		if(oldContentFilter==null)
			println(contentFilter.getClass().getName()+" registered "
				+"for "+acceptedContenttype.toLowerCase()
				+" type.");
		else
			println(contentFilter.getClass().getName()+" replaced "
				+oldContentFilter.getClass().getName()
				+"for "+acceptedContenttype.toLowerCase()
				+" type.");
	}

	/**
	 * Attempts to get a content from a given input stream.
	 * The ContentManager attempts to select an appropriate content filter from
	 * the set of registered content filter.
	 *
	 * @param InputStream an input stream
	 * @return a Content from the input stream
	 * @exception IOException if an input stream access error occurs
	 */
	public static Content getContent(URLConnection urlConnection) throws IOException
	{
		if(urlConnection==null)return null;
		println("ContentManager.getContent(\"" + urlConnection + "\")");
		Content content=getContent(urlConnection.getContentType(),urlConnection.getInputStream());
		if(content==null)//Unknown content type
			//Create a content which contain the file name of this URL
		{
			content=new Content();
			content.title.add(urlConnection.getURL().getFile());
			println("		Only return: " + content.title.get(0));
		}
		return content;
	}
	/**
	 * Attempts to get a content from a given input stream .
	 * The ContentManager attempts to select an appropriate content filter from
	 * the set of registered content filter.
	 *
	 * @param String the content type of the input stream that the InputStream references, or null if not known.
	 * @param InputStream an input stream 
	 * @return a Content to the input stream 
	 * @exception IOException if a input stream access error occurs
	 */
	private static Content getContent(String contentType,
			InputStream in) throws IOException
	{
		if(contentType==null)
		{
			println("	Content type: unknown");
			return null;
		}
		else
		{
			println("	Content type: " + contentType);
			ContentFilter contentFilter=(ContentFilter)contentFilters.get(contentType.toLowerCase());//In fact, contentType should always without any upper letter.
			if(contentFilter==null)
			{
				int index=contentType.indexOf('/');
				if(index>0)//In fact, index should always >0
					contentFilter=(ContentFilter)contentFilters.get(contentType.toLowerCase().substring(0,index));//Maybe we have appointed a type for this subtype.
				if(contentFilter==null)
				{
					println("	No suitable content filter for "+contentType+" type.");
					return null;
				}
			}
			if(in==null)
				return null;
			else
				return contentFilter.getContent(in);
		}
	}


	/**
	 * Registers the content processor with the ContentManager.
	 *
	 * @param robot the robot see after document data on intranet or internet
	 * @param contentProcessor the new content processor that is to be registered with the ContentManager
	 */
	private static java.util.Hashtable contentProcessors = new java.util.Hashtable(5);
	public static synchronized void registerContentProcessor(Robot robot,ContentProcessor contentProcessor)
	{	
		ContentProcessor oldContentProcessor=(ContentProcessor)contentProcessors.put(robot,contentProcessor);
		if(oldContentProcessor==null)
			println(contentProcessor.getClass().getName()+" registered "
				+"for "+robot+".");
		else
			println(contentProcessor.getClass().getName()+" replaced "
				+oldContentProcessor.getClass().getName()
				+"for "+robot+".");
	}

	/**
	 * Sort and Index a content object.
	 *
	 * @return a wordID array of all keywords from the content
	 *
	 * @param obj an robot object
	 * @param content an Content object
	 * @exception SQLException if an index database access error occurs
	 */
	public static int[] getWordIDs(Object obj, Content content)throws SQLException
	{
		if(obj==null)
		{
			println("A null value for obj in zyh.net.ContentManager.getWordIDs(Object obj, Content content)");
			return new int[0];
		}
		println(obj+" call ContentManager.getWordIDs for urlID" +content.urlID+ ".");
		ContentProcessor contentProcessor=(ContentProcessor)contentProcessors.get(obj);
		if(contentProcessors==null)
		{
			println("	No registered content processor for "+obj+".");
			return new int[0];
		}
		try
		{
			return contentProcessor.getWordIDs(content);
		}
		catch(SQLException se)
		{
			se.printStackTrace();
			println("java.sql.SQLException:"+se);
			throw se;
		}
	}
}
