/*
 *
 *  @author  Zhao Yonghong (zhaoyh@bigfoot.com)
 *
 *	Copyright 2000 by HXKM Inc., 9 Station Road, Xiangtan, 411100, P.R. China
 *	All rights reserved.
 * 
 *  You can distribute freely it without modification.
 *  You also can modify and distribute it but please keep our statements unchanged.
 *  For more licences conditions, please refer to
 *  <A HREF="http://www.fsf.org/copyleft/gpl.html">GNU General Public License (GPL)</A>
 *	or
 *  <A HREF="http://language.perl.com/misc/Artistic.html">Artistic License</A>
 *
 */

package zyh.robot;

import java.util.*;

import java.sql.*;

import java.io.*;

/**
 * Uses ContentSorter to extract the keywords from Content object,
 * and maintain the keywords index dabase.
 *
 * This ContentProcessor will filter stopword(noise word) and
 * calculate the weight according to its quoted frequency and quoted location.
 * The incorporate of synonyms, plural and p.p are not complemented.
 *
 * @see ContentProcessor 
 * @see Content 
 */
public class ContentSorter implements ContentProcessor
{
	//The JDBC Connection for the keyword index database
	private	Connection connection;

	private int keywordID;//The unique ID for every keyword
	private Hashtable keywordTable;
	private Hashtable stopwordTable;

	//A ResultSet.CONCUR_UPDATABLE select prepared statement
	private	PreparedStatement selectPreparedStatement;
	//the parameterIndex in selectPreparedStatement
	private static final int WORDID_PARA=1;

	//A CONCUR_UPDATABLE resultset
	private ResultSet updatableResultSet=null;
	//the columnIndex in updatableResultSet
	private static final int KEYWORD=1;
	private static final int WORDID=2;
	private static final int URLS=3;

	/**
	 * Creates a ContentSorter object
	 *
	 * @param jdbcurl the jdbcurl for the stopword databse and keyword database
	 * @param keywordTableName the table name which contains all keywords
	 * @param stopwordTableName the table name which contains all stopwords
	 */	
	public ContentSorter(String jdbcurl,
		String keywordTableName,
		String stopwordTableName)throws SQLException
	{
		connection = DriverManager.getConnection(jdbcurl, "", "");
		Statement stmt = connection.createStatement();
		stmt.setFetchSize(100);

		ResultSet rs=stmt.executeQuery("select max(wordID) from "+keywordTableName);
		keywordID=1;
		if(rs.next())
			keywordID=rs.getInt(1)+1;
		rs.close();

		keywordTable=new Hashtable(keywordID+1000);

		rs=stmt.executeQuery("select keyword,wordID from "+keywordTableName);
		while (rs.next())
		{
			keywordTable.put(rs.getString(1),new Integer(rs.getInt(2)));
		}
		rs.close();
		
		stopwordTable=new Hashtable(600);

		rs=stmt.executeQuery("select stopword from "+stopwordTableName);
		String stopword;
		while (rs.next())
		{
			stopword = rs.getString(1);
			stopwordTable.put(stopword,stopword);
		}
		rs.close();
		stmt.close();

		selectPreparedStatement= connection.prepareStatement(
			"select keyword,wordID,urls from "+keywordTableName
			+" where wordID=?",ResultSet.TYPE_SCROLL_SENSITIVE,
			ResultSet.CONCUR_UPDATABLE);
	}
	/**
	 * Uses to release all occupied resource by this content processor
	 */
	public synchronized void close()throws SQLException
	{
		//Uses to close the JDBC connection
		if(connection!=null)
		{
			if(updatableResultSet!=null)updatableResultSet.close();
			selectPreparedStatement.close();
			connection.close();
			connection=null;
		}
	}

	/*
	 * Add a new URLWeight object into index database
	 */
	private synchronized void addURLWeight(URLWeight uw)throws SQLException
	{
		if(updatableResultSet!=null)updatableResultSet.close();
		
		selectPreparedStatement.setInt(WORDID_PARA,uw.wordID);
		updatableResultSet=selectPreparedStatement.executeQuery();
		if(updatableResultSet.next())
		{
			URLs urls=(URLs)updatableResultSet.getObject(URLS);
			urls.urlWeights.add(uw);
			updatableResultSet.updateObject(URLS,urls);
			updatableResultSet.updateRow();
		}
		else
			throw new SQLException("Error: there is an unknown error when distributing wordID.");
	}

	/*
	 * Insert a new keyword into index database
	 */
	private synchronized void insertKeyword(int wordID,String keyword)throws SQLException
	{
		if(updatableResultSet==null)
		{
			selectPreparedStatement.setInt(WORDID_PARA,wordID);
			updatableResultSet=selectPreparedStatement.executeQuery();
			if(updatableResultSet.next())
				throw new SQLException("Error: there is an unknown error when distributing wordID"+wordID+".");
		}
		updatableResultSet.moveToInsertRow();
		updatableResultSet.updateString(KEYWORD,keyword);
		updatableResultSet.updateInt(WORDID,wordID);
		updatableResultSet.updateObject(URLS,new URLs());
		updatableResultSet.insertRow();
	}

	/**
	 * Sort and Index a content object.
	 *
	 * @return a wordID array of all keywords from the content
	 *
	 * @param content an Content object
	 * @exception SQLException if an index database access error occurs
	 */
	public int[] getWordIDs(Content content)throws SQLException
	{
		try
		{
			Hashtable wordsTable=new Hashtable();//A table which contains all indexed keywords and their weight

			sortWords(wordsTable,content.urlID,content.title,1000,3000);//The keywords in the title have the largest weight
			sortWords(wordsTable,content.urlID,content.keywords,200,600);
			sortWords(wordsTable,content.urlID,content.summary,40,120);
			sortWords(wordsTable,content.urlID,content.text,1,600);//The keywords in the text have the smallest weight
			int[] wordIDs=new int[wordsTable.size()];
			int keywordNumber=wordIDs.length;
			int nCount=0;
			Enumeration words = wordsTable.elements();

			URLWeight uw;
			while(words.hasMoreElements())
			{
				uw=(URLWeight) words.nextElement();

				//Change wordWeight from an absolute value to a relative value
				uw.wordWeight=uw.wordWeight*10000/keywordNumber;

				wordIDs[nCount]=uw.wordID;
				nCount++;

				addURLWeight(uw);				
			}

			wordsTable.clear();
			return wordIDs;
		}
		catch(Exception e)
		{
			e.printStackTrace();
			System.out.println(e.getMessage());
//			System.exit(1);
			return null;
		}
	}

	/*
	 * @param wordsTable Use it to keep all indexed keywords and their weight
	 * @param urlID the url ID of the url which contains the relative content
	 * @param ll this LinkedList object in Content component objects
	 * @param wordWeight the weight of every keyword
	 * @param maxwordWeight Use it to avoid fooling by the abusive use of keyword metatags
	 */
	private void sortWords(Hashtable wordsTable,int urlID,LinkedList ll,int wordWeight,int maxwordWeight)throws SQLException
	{
		if(ll==null)return;
		Object obj;
		for(int i=0;i<ll.size();i++)
		{
			obj=ll.get(i);
			if(obj instanceof String)
				sortWords(wordsTable,urlID,(String)obj,wordWeight,maxwordWeight);
			else if(obj instanceof LinkedList)
				sortWords(wordsTable,urlID,(LinkedList)obj,wordWeight,maxwordWeight);
			else
				throw new SQLException("Unexpect class: "+obj.getClass().getName()+" in sortWords(int urlID,LinkedList ll,int wordWeight)");
		}
	}
	/*
	 * @param wordsTable Use it to keep all indexed keywords and their weight
	 * @param urlID the url ID of the url which contains the relative content
	 * @param text a String object which contain some words.
	 * @param wordWeight the weight of every keyword
	 * @param maxwordWeight Use it to avoid fooling by the abusive use of keyword metatags
	 */
	private void sortWords(Hashtable wordsTable,int urlID,String text,int wordWeight,int maxwordWeight)throws SQLException
	{
		int last=0;
		int current=0;
		while(last<text.length())
		{
			for(current=last;current<text.length();current++)
			{
				if(!Character.isLetter(text.charAt(current))
					|| text.charAt(current)>255//Omit the Unicode character
					)
					break;
			}
			if(current>last+1)//A keyword has two letters at least
			{
				String currentWord=text.substring(last,current).toLowerCase();
				if(!stopwordTable.containsKey(currentWord))//Omit noise word
				{
					Object obj=keywordTable.get(currentWord);
					int currentWordID;
					if(obj==null)//A new keyword
					{
						currentWordID=keywordID;
						keywordID++;
						keywordTable.put(currentWord,new Integer(currentWordID));

						insertKeyword(currentWordID,currentWord);
					}
					else
						currentWordID=((Integer)obj).intValue();
					obj=wordsTable.get(currentWord);
					if(obj==null)
						wordsTable.put(currentWord,new URLWeight(urlID,currentWordID,wordWeight));
					else if(((URLWeight)obj).wordWeight<maxwordWeight)
							((URLWeight)obj).wordWeight+=wordWeight;
				}
			}
			last=current+1;
		}
	}

	/**
	 * Search.java use this function to sort the url list
	 * which is gotten from OR operator.
	 *
	 * @return a sorted urlID array according to their word weight
	 *
	 * @param urlsColumnIndex the index of a column which contains URLs object 
	 * @param resultSet a resultset contains "urls" column at least
	 */	
	public static final int[] sortURLs(int urlsColumnIndex,ResultSet resultSet)throws SQLException
	{
		Hashtable uwsTable=new Hashtable(500);

		URLs urls;
		URLWeight uw,uwtemp;
		while (resultSet.next())
		{
			urls=(URLs)resultSet.getObject(urlsColumnIndex);
			if(urls!=null)
			{
				for(int i=0;i<urls.urlWeights.size();i++)
				{
					uw=(URLWeight)urls.urlWeights.get(i);

					uwtemp=(URLWeight)uwsTable.get(new Integer(uw.urlID));
					if(uwtemp==null)
						uwsTable.put(new Integer(uw.urlID),uw);
					else
						uwtemp.wordWeight+=uw.wordWeight;
				}
			}
		}
		URLWeight[] urlWeights=new URLWeight[uwsTable.size()];
		int nCount=0;
		Enumeration uws = uwsTable.elements();
		while(uws.hasMoreElements())
		{
			urlWeights[nCount]=(URLWeight) uws.nextElement();
			nCount++;
		}
		Arrays.sort(urlWeights);
		int[] urlIDs=new int[urlWeights.length];
		for(int i=0;i<urlIDs.length;i++)
			urlIDs[i]=urlWeights[i].urlID;
		return urlIDs;
	}

	/**
	 * Search.java use this function to extract the url list
	 * which is gotten from AND operator.
	 *
	 * @return a sorted urlID array according to their word weight
	 *
	 * @param urlsColumnIndex the index of a column which contains URLs object 
	 * @param resultSet a resultset contains "urls" column at least
	 * @param sqls some sql commands 
	 * @param statement a statement
	 */	
	public static final int[] sortURLs(int urlsColumnIndex,ResultSet resultSet,
		String[] sqls,
		Statement statement)throws SQLException
	{
		if(!resultSet.next())
		{
			resultSet.close();
			return new int[0];
		}
		//Only one row in the resultSet.
		URLs urls=(URLs)resultSet.getObject(urlsColumnIndex);
		resultSet.close();
		URLWeight[] uws=new URLWeight[urls.urlWeights.size()];
		for(int i=0;i<uws.length;i++)
			uws[i]=(URLWeight)urls.urlWeights.get(i);

		int nCount=uws.length;

		URLWeight uw;
		for(int i=0;i<sqls.length;i++)
		{
			resultSet=statement.executeQuery(sqls[i]);
			if(!resultSet.next())
			{
				resultSet.close();
				return new int[0];
			}
			urls=(URLs)resultSet.getObject(urlsColumnIndex);
				resultSet.close();
			for(int j=0;j<urls.urlWeights.size();j++)
			{
				uw=(URLWeight)urls.urlWeights.get(j);
				for(int k=0;k<nCount;k++)
				{
					if(uws[k].urlID==uw.urlID)
					{
						uws[k].wordID=-1;
						break;
					}
				}
			}
			int temp=0;
			for(int j=0;j<nCount;j++)
			{
				if(uws[j].wordID!=-1)
					uws[j].wordWeight=-1;//Set a lowest word weight so that it will be placed at the end after sort operation
				else
				{
					temp++;
					uws[j].wordID=1;
				}
			}
			if(temp==0)return new int[0];
			Arrays.sort(uws,0,nCount);
			nCount=temp;
		}
		int[] urlIDs=new int[nCount];
		for(int i=0;i<urlIDs.length;i++)
			urlIDs[i]=uws[i].urlID;
		return urlIDs;
	}
}

/**
 * An auxillary class which holds the weight information related
 * to a specific kewword.
 */
class URLWeight implements java.lang.Comparable
{
	int urlID;
	int wordID;
	int wordWeight;

	public URLWeight(int urlID,int wordID,int wordWeight)
	{
		this.urlID=urlID;
		this.wordID=wordID;
		this.wordWeight=wordWeight;
	}
	public int compareTo(Object o)
	{
		URLWeight uw=(zyh.robot.URLWeight)o;
		//Rem:the bigger wordWeight should be placed in front
		return (wordWeight>uw.wordWeight ? -1 : (wordWeight==uw.wordWeight ? 0 : 1));
	}
}
/**
 * An auxillary class which holds URLWeight information related
 * to a keyword.
 */
class URLs implements java.io.Serializable
{
	transient ArrayList urlWeights;
	public URLs()
	{
		urlWeights=new ArrayList(5);
	}
	public URLs(URLWeight uw)
	{
		urlWeights=new ArrayList(5);
		urlWeights.add(uw);
	}

	private void writeObject(ObjectOutputStream out)throws IOException
	{
		URLWeight[] uws=new URLWeight[urlWeights.size()];
		for(int i=0;i<uws.length;i++)
			uws[i]=(URLWeight)urlWeights.get(i);
		Arrays.sort(uws);

		out.writeInt(uws.length);
		for(int i=0;i<uws.length;i++)
		{
			out.writeInt(uws[i].urlID);
			out.writeInt(uws[i].wordWeight);
		}
	}

	private void readObject(ObjectInputStream in)throws IOException,ClassNotFoundException
	{
		int size=in.readInt();
		urlWeights=new ArrayList(size);
		for(int i=0;i<size;i++)
			urlWeights.add(new URLWeight(in.readInt(),0,in.readInt()));
	}
}