/*
 *
 *  @author  Zhao Yonghong (zhaoyh@bigfoot.com)
 *
 *	Copyright 2000 by HXKM Inc., 9 Station Road, Xiangtan, 411100, P.R. China
 *	All rights reserved.
 * 
 *  You can distribute freely it without modification.
 *  You also can modify and distribute it but please keep our statements unchanged.
 *  For more licences conditions, please refer to
 *  <A HREF="http://www.fsf.org/copyleft/gpl.html">GNU General Public License (GPL)</A>
 *	or
 *  <A HREF="http://language.perl.com/misc/Artistic.html">Artistic License</A>
 *
 */

package zyh.robot;

import java.net.*;
import java.io.*;
import java.util.*;

import zyh.util.ThreadPool;//A thread pool class

import zyh.net.http.HttpURLConnection;

/**
 * A robot can parse the content of a URL and
 * extract all links, then parse those new URLs, recursively.
 *
 * If you are behind a firewall, please read also
 * <a href=http://www.mokabyte.it/2000/06/firewallutil.htm>Let Java Applications Through Firewall</a>
 */
public class Robot implements Runnable
{
	//The URL manager which decides on the visit policy
	private URLManager urlManager;

	//The thread pool manager which manages the thread distribution
	private	ThreadPool threadPool;

	//Requests 1 document per requestInterval milliseconds
	private long requestInterval;
	
	private	InetAddress proxyInetAddress;
	private int proxyPort;

	private static int robotCounter=0;
	private int robotID=0;
	/**
	 * Creates a Robot object
	 *
	 * @param requestInterval Requests 1 document per n milliseconds
	 * Suggestion: Using 60000~300000 milliseconds to avoid the potential overload of the destination host
	 * Warning: Using a smaller value for requestInterval only when you're exploring your own website
	 * @param urlManager The URL manager which decides on the visit policy
	 * @param threadPool The thread pool manager which manages the thread distribution
	 * @param proxyInetAddress the InetAddress of HTTP proxy
	 * @param proxyPort the port of HTTP proxy
	 */

	public Robot(long requestInterval,
		URLManager urlManager,
		ThreadPool threadPool,
		InetAddress proxyInetAddress,int proxyPort
		)
	{
		this.requestInterval=requestInterval;
		if(requestInterval<0)requestInterval=0;

		this.urlManager=urlManager;
		this.threadPool=threadPool;

		this.proxyInetAddress=proxyInetAddress;
		this.proxyPort=proxyPort;

		robotCounter++;
		robotID=robotCounter;

		threadPool.run(this);
	}

	public String toString()
	{
		return "zyh.robot.Robot"+robotID;
	}

	private Thread myThread=null;
	public final boolean isAlive()
	{return myThread!=null;}
	private boolean idleFlag=false;
	public final boolean isIdle()
	{return idleFlag|| myThread==null;}
	public void stop()
	{myThread=null;}

	public void run()
	{		
		myThread=Thread.currentThread();

		URLInfo currentURLInfo;
		URLConnection connection;
		String cookie;
		Content content;
		while(myThread!=null)
		{
			if((currentURLInfo=urlManager.getNextUnprocessedURLInfo())==null)//Robot is on call for requesting the available retrieval task from URL manager's queue.
			{
				try
				{
					idleFlag=true;
					myThread.sleep(5000);
				}
				catch(InterruptedException e){}
				continue;
			}
			idleFlag=false;
			if(requestInterval>0)//Requests 1 document per requestInterval milliseconds
			{
				try
				{
					myThread.sleep(requestInterval);
				}
				catch(InterruptedException e){}
			}
			try
			{
				content=null;

				if(!currentURLInfo.url.getProtocol().equals("http"))
					//The protocol of URL format is not HTTP.
					//This robot does support other protocols, such as "file" protocol.
				{
					connection=currentURLInfo.url.openConnection();
					connection.setAllowUserInteraction(false);
					if(!connection.getDoInput())connection.setDoInput(true);

					connection.connect();

					currentURLInfo.responseCode =HttpURLConnection.HTTP_OK;
					currentURLInfo.responseMessage = "OK";

					if(currentURLInfo.urlType==URLManager.INTERESTED_URL)
						content=ContentManager.getContent(connection);
				}
				else
				{
					connection=new HttpURLConnection(currentURLInfo.url,urlManager.getMailbox(),proxyInetAddress,proxyPort);
					cookie=urlManager.getCookie(currentURLInfo.url.getHost());
					if(cookie!=null)
						connection.setRequestProperty("Cookie",cookie);
					if(currentURLInfo.referers!=null && currentURLInfo.referers.size()>0)
						connection.setRequestProperty("Referer",currentURLInfo.referers.get(0).toString());
					
					connection.connect();

					currentURLInfo.responseCode =((HttpURLConnection)connection).getResponseCode();
					currentURLInfo.responseMessage = ((HttpURLConnection)connection).getResponseMessage();
					if(connection.getHeaderField("Set-Cookie")!=null)
					{
						urlManager.setCookie(currentURLInfo.url.getHost(),connection.getHeaderField("Set-Cookie"));
					}

					if(currentURLInfo.urlType==URLManager.INTERESTED_URL)
						content=ContentManager.getContent(connection);

					((HttpURLConnection)connection).disconnect();
				}

				if(content!=null)
				{
					for(int i=0;i<content.links.length;i++)
					{
						for(int j=0;j<content.links[i].size();j++)
						{
							try
							{
								urlManager.addURL(new URL(connection.getURL(),//Not currentURLInfo.url because of HTTP_MOVED_PERM and HTTP_MOVED_TEMP
									(String)content.links[i].get(j)),i,currentURLInfo);
							}
							catch(java.net.MalformedURLException mue)
							{
								//This url is with an invalid format
								urlManager.println(mue+":"+content.links[i].get(j)+" in "+currentURLInfo.url);
							}
						}
					}

					content.urlID=currentURLInfo.urlID;

					int[] wordIDs=ContentManager.getWordIDs(this, content);

					urlManager.setKeywords(currentURLInfo.urlID,wordIDs);

					currentURLInfo.keywordNumber=wordIDs.length;

				}
			}
			catch (Exception e)
			{
				currentURLInfo.responseCode=URLManager.UNKNOWN_LINK;
				currentURLInfo.responseMessage=""+e;
			}
		}
	}
}