/*
 *
 *  @author  Zhao Yonghong (zhaoyh@bigfoot.com)
 *
 *	Copyright 2000 by HXKM Inc., 9 Station Road, Xiangtan, 411100, P.R. China
 *	All rights reserved.
 * 
 *  You can distribute freely it without modification.
 *  You also can modify and distribute it but please keep our statements unchanged.
 *  For more licences conditions, please refer to
 *  <A HREF="http://www.fsf.org/copyleft/gpl.html">GNU General Public License (GPL)</A>
 *	or
 *  <A HREF="http://language.perl.com/misc/Artistic.html">Artistic License</A>
 *
 */

package zyh.robot;

import java.io.*;
import java.util.*;

import java.net.URLConnection;
import java.net.URL;
import java.net.InetAddress;

import java.sql.*;

import zyh.net.http.HttpURLConnection;

/**
 * The basic service for maintaining a list of visited URLs
 * and tell Robot how to deal with a specific URL.
 *
 * To avoid revisiting the same URL over and over, a list of
 * visited URLs should be maintained in memory or database. 
 * The zyh.robot.URLManager uses a hash table to hold all 
 * visited URLs and their related information. It decides 
 * whether to visit a URL or not according to the specific
 * interested URL identifiers, the maximal recursive depth
 * and the Robots Exclusion Protocol.
 *
 * If you wish to modify it for a large url database,
 * you can not load all url information into memory from
 * the url index datbase. You should seek the specific keyword
 * or url through JDBC drvier one by one.
 *
 */
public class URLManager
{
	/**
	 * Sets the logging/tracing Writer that is used by the ContentFilterManager
	 * and all ContentFilter.
	 *
	 * @param out the new logging/tracing PrintWriter; to disable, set to null
	 */
	private java.io.PrintWriter logWriter = null;
	public synchronized void setLogWriter(java.io.PrintWriter out)
	{
		logWriter = out;
	}

	/**
	 * Gets the log writer.	
	 */
	public java.io.PrintWriter getLogWriter()
	{
		return logWriter;
	}

	/**
	 * Prints a message to the current ContentFilter log writer.
	 *
	 * @param message a log or tracing message
	 */
	public synchronized void println(String message)
	{
		if (logWriter != null)
		{
			logWriter.println(message);
			logWriter.flush();
		}
	}

	private int maxDepth;//the maximal recursive depth
	private String[] interestedIdentifiers=null;//

	//Please provide your email so that server maintainers can contact you in case of problems.
	private String mailbox=null;
	/**
	 * return the email of robot master
	 */
	protected String getMailbox()
	{return mailbox;}

	/* Http Proxy Parameters */
	private InetAddress proxyInetAddress=null;
	private int proxyPort=8080;

	// The LinkedList array containing links that haven't been processed yet.
	private LinkedList[] unprocessedURLsLists;

	//The JDBC Connection for the url index database
	private	Connection connection;

	private int urlID;//The unique ID for every url

	/*
	 * The hashtable containing all known links.
	 * It's possible to load all known links into memory
	 * since we're constructed a full-text search
	 * engine for the limited hosts or recursive depth.
	 */
	private Hashtable urlsTable=null;

	//A ResultSet.CONCUR_UPDATABLE select prepared statement
	private	PreparedStatement selectPreparedStatement;
	//the parameterIndex in selectPreparedStatement
	private static final int URLID_PARA=1;
	//A CONCUR_UPDATABLE resultset
	private ResultSet updatableResultSet=null;
	//the columnIndex in updatableResultSet
	private static final int URL=1;
	private static final int URLID=2;
	private static final int URLTIME=3;//It's used to refresh url page. Not implemented.
	private static final int KEYWORDS=4;

	/**
	 * Creates a URLManager object
	 *
	 * @param maxDepth the maximal recursive depth
	 * @param interestedIdentifiers all interested indentifiers
	 * @param mailbox please provide your mailbox so that server maintainers can contact you in case of problems
	 * @param proxyInetAddress the InetAddress of HTTP proxy
	 * @param proxyPort the port of HTTP proxy
	 * @param jdbcurl the JDBC url of index database
	 * @param urlsTableName the table name which contains all urls
	 */
	public URLManager(int maxDepth,
		String[] interestedIdentifiers,
		String mailbox,
		InetAddress proxyInetAddress,int proxyPort,
		String jdbcurl,
		String urlsTableName)throws SQLException
	{
		if(maxDepth < 1)maxDepth =1;
		this.maxDepth=maxDepth;
		this.interestedIdentifiers=interestedIdentifiers;
		this.mailbox=mailbox;
		this.proxyInetAddress=proxyInetAddress;
		this.proxyPort=proxyPort;

		unprocessedURLsLists = new LinkedList[Content.LINK_KIND];
		unprocessedURLsLists[Content.BACKGROUND_LINK] = new LinkedList();
		unprocessedURLsLists[Content.SRC_LINK] = new LinkedList();
		unprocessedURLsLists[Content.HREF_LINK] = new LinkedList();

		connection = DriverManager.getConnection(jdbcurl,"","");
		Statement stmt = connection.createStatement();
		stmt.setFetchSize(100);

		ResultSet rs=stmt.executeQuery("select max(urlID) from "+urlsTableName);
		urlID=1;
		if(rs.next())
			urlID=rs.getInt(1)+1;
		rs.close();

		urlsTable=new Hashtable(urlID+200);

		rs=stmt.executeQuery("select url,urlID from "+urlsTableName);
		while (rs.next())
		{
			try
			{
				urlsTable.put(new URL(rs.getString(1)),new Integer(rs.getInt(2)));
			}
			catch(java.net.MalformedURLException mue)
			{}
		}
		rs.close();
		
		selectPreparedStatement= connection.prepareStatement(
			"select url,urlID,urltime,keywords from "+urlsTableName
			+" where URLID=?",ResultSet.TYPE_SCROLL_SENSITIVE,
			ResultSet.CONCUR_UPDATABLE);
	}

	/*
	 * insert a new url into url database
	 */
	private synchronized void insertURL(int urlID,String url)throws SQLException
	{
		if(updatableResultSet==null)
		{
			selectPreparedStatement.setInt(URLID_PARA,urlID);
			updatableResultSet=selectPreparedStatement.executeQuery();
			if(updatableResultSet.next())
				throw new SQLException("Error: there is an unknown error when distributing urlID"+urlID+".");
		}
		updatableResultSet.moveToInsertRow();
		updatableResultSet.updateString(URL,url);
		updatableResultSet.updateInt(URLID,urlID);
		updatableResultSet.updateTime(URLTIME,new Time(System.currentTimeMillis()));
		updatableResultSet.updateObject(KEYWORDS,new WordIDs());
		updatableResultSet.insertRow();
	}

	/*
	 * The Robots Exclusion Protocol
	 * http://info.webcrawler.com/mak/projects/robots/exclusion.html
	 *
	 * Uses parseRobot(URL url) to load the disallow list in robots.txt
	 * before requesting the referenced URL.
	 */
	private Hashtable disallowableURLs=new Hashtable();
	private synchronized void parseRobot(URL url)
	{
		//the protocol of URL format is not HTTP
		if(!url.getProtocol().equals("http"))return;

		String host=url.getHost();

		LinkedList ll=(LinkedList)disallowableURLs.get(host);
		if(ll!=null)return;//Has visited "http://"+ host +"/robots.txt"
		ll=new LinkedList();
		disallowableURLs.put(host,ll);		
		try
		{
			//retrieve and parse robots.txt
			HttpURLConnection connection=new HttpURLConnection(new URL("http://"+ host +"/robots.txt"),mailbox,proxyInetAddress, proxyPort);

			connection.connect();

			BufferedReader br= new BufferedReader(new InputStreamReader(connection.getInputStream()));
			String line;
			while((line=br.readLine())!=null)
			{
				if(line.indexOf("Disallow: ")==0 && line.length()>10)
					ll.add("http://"+ host +line.substring(10));
			}
		}
		catch(Exception e)
		{
			println("Load "+host+"'s robots.txt Error: "+e);
		}
	}

	/**
	 * interested URL type
	 * the URL should be visited and
	 * Robot should extract all referenced URL links from this URL
	 */
	public static final int INTERESTED_URL=10001;
	/**
	 * referenced URL type
	 * This url is on other friendly sites
	 * The URL should be visited but not filtered the content
	 */
	public static final int REFERENCED_URL=10002;
	/**
	 * Untouched URL type
	 * The depth of this url beyond the maximal recursive depth
	 * The URL should be visited but not filtered the content
	 */
	public static final int UNTOUCHED_URL=10003;
	/**
	 * Adds a destination URL
	 *
	 * Checks if a URL is in the list of known URL's, and if it's not, adds
	 * it to the list.
	 *
	 * @param baseURL the start url
	 */
	public void addURL(URL baseURL)throws SQLException
	{
		addURL(baseURL,Content.HREF_LINK,null);
	}
	/**
	 * Adds a destination URL
	 *
	 * Checks if a URL is in the list of known URL's, and if it's not, adds
	 * it to the list.
	 *
	 * @param url the url which will be processed
	 * @param linkType the url link type in the processed document
	 * @param parentURLInfo the URLInfo object which contains this urlInfo
	 */
	public synchronized void addURL(URL url,int linkType,URLInfo parentURLInfo)throws SQLException
	{	
		Object obj=urlsTable.get(url);
		URLInfo urlInfo; 
		if(obj==null || !(obj instanceof URLInfo))
		{
			urlInfo = new URLInfo(url);
			urlInfo.referers = new LinkedList();
			if(parentURLInfo==null)
				urlInfo.depth=1;
			else
			{
				urlInfo.depth=parentURLInfo.depth+1;
				urlInfo.referers.add(parentURLInfo.url);
			}
			if(obj==null)
				urlInfo.urlType=getURLType(url,urlInfo.depth);
			else
			{
				urlInfo.urlType=VISITED_URL;
				if(obj instanceof Integer)
					urlInfo.urlID=((Integer)obj).intValue();
				else//It should not occur if you haven't misused urlsTable.put(key,value);
					throw new SQLException("A programming error in URLManager.java");
			}
			addURL(urlInfo,linkType);
		}	
		else
		{
			urlInfo=(URLInfo)obj;
			boolean isTouched=(urlInfo.depth <= maxDepth);
			if(parentURLInfo==null)
				urlInfo.depth=1;
			else
			{
				if(urlInfo.depth>parentURLInfo.depth+1)
					urlInfo.depth=parentURLInfo.depth+1;
				if(!urlInfo.referers.contains(parentURLInfo.url))
					urlInfo.referers.add(parentURLInfo.url);
			}
			if(!isTouched && urlInfo.depth <= maxDepth)
			{
				if(urlInfo.urlType==UNTOUCHED_URL)
					urlInfo.urlType=INTERESTED_URL;
				unprocessedURLsLists[linkType].add(urlInfo);

				insertURL(urlInfo.urlID,urlInfo.url.toString());

			}
		}
	}
	/*
	 * Gets the url processed type
	 *
	 * @param url the url which will be processed
	 * @param depth the recursive depth which will be processed
	 */
	private synchronized int getURLType(URL url,int depth)throws SQLException
	{	
		parseRobot(url);

		LinkedList ll=(LinkedList)disallowableURLs.get(url.getHost());
		if(ll!=null && ll.size()>0)
		{
			String spec=url.toString();
			for(int i=0;i<ll.size();i++)
			{
				if(spec.indexOf((String)ll.get(i))==0)
					return DISALLOWABLE_LINK;
			}
		}

		int urlType;
		if(interestedIdentifiers==null)
			urlType=INTERESTED_URL;
		else
		{
			urlType=REFERENCED_URL;
			String spec=url.toString();
			for(int i=0;i<interestedIdentifiers.length;i++)
			{
				if(spec.indexOf(interestedIdentifiers[i])>=0)
				{
					urlType=INTERESTED_URL;
					break;
				}
			}
		}
		if(urlType==INTERESTED_URL
			&& depth > maxDepth)//the depth of this url beyond the maximal recursive depth
			return UNTOUCHED_URL;
		return urlType;
	}

	/*
	 * Adds a processed URLInfo object
	 *
	 * @param urlInfo the processed URLInfo object
	 * @param linkType the url link type in the processed document
	 */
	private synchronized int addURL(URLInfo urlInfo,int linkType)throws SQLException
	{
		if(urlInfo.urlType!=VISITED_URL)
		{
			urlInfo.urlID=urlID;
			println("Use "+urlID+" as the urlID of "+urlInfo);
			urlID++;
		}
		else
			println("The assigned urlID of "+urlInfo+" is "+urlInfo.urlID);

		urlsTable.put(urlInfo.url,urlInfo);

		if((urlInfo.urlType==URLManager.INTERESTED_URL
			|| urlInfo.urlType==REFERENCED_URL)
			&& urlInfo.depth <= maxDepth
			)
		{
			unprocessedURLsLists[linkType].add(urlInfo);

			insertURL(urlInfo.urlID,urlInfo.url.toString());
		}
		return urlID;
	}

	/**
	 * Set the keyword list of a destination URL
	 *
	 * @param urlID the urlID which will be processed
	 * @param wordIDs the wordIDs array which contains this keywords for the url which alloted thie specific urlID.
	 */
	public synchronized void setKeywords(int urlID,int[] wordIDs)throws SQLException
	{	
		if(updatableResultSet!=null)updatableResultSet.close();
		selectPreparedStatement.setInt(URLID_PARA,urlID);
		updatableResultSet=selectPreparedStatement.executeQuery();
		if(updatableResultSet.next())
		{
			if(wordIDs.length>0)
			{
				updatableResultSet.updateObject(KEYWORDS,new WordIDs(wordIDs));
				updatableResultSet.updateRow();
				println("Indexed "+wordIDs.length+(wordIDs.length>1?" keywords":"keyword")
					+" for "+updatableResultSet.getObject(URL)+".");			}
			else
				println("Indexed nothing for "+updatableResultSet.getObject(URL)+".");
		}
		else
			throw new SQLException("Error: Failed to find the url which contians a urlID of "+urlID+" .");
	}
	private Hashtable cookies = new Hashtable(20);
	/**
	 * Gets a cookie for the specific host
	 */
	public String getCookie(String host)
	{
		return (String)cookies.get(host);
	}
	/**
	 * Sets a cookie for the specific host
	 */
	public void setCookie(String host,String cookie)
	{
		cookies.put(host,cookie);
	}
	
	/**
	 * Gets an unprocessed URLInfo object from the waiting queue.
	 *
	 * If you wish to limit that no more than one robot visits the same host at the same time
	 * when you're using a lot of robots, you should modify this function to
	 * public URLInfo getNextUnprocessedURLInfo(Robot robot,URL latestVisitedURL) and
	 * maintain a map of every robot's latest visited host.
	 * You can also using rotate queries between different servers
	 * in a round-robin fashion.
	 */
	public URLInfo getNextUnprocessedURLInfo()
	{
		for(int i=0;i<unprocessedURLsLists.length;i++)
		{
			if(!unprocessedURLsLists[i].isEmpty())
			{
				URLInfo urlInfo= (URLInfo) unprocessedURLsLists[i].get(0);
				unprocessedURLsLists[i].remove(0);
				return urlInfo;
			}
		}
		return null;
	}

	/**
	 * Outputs an html format report which contains the link information.
	 * @param out a PrintWriter output object.
	 */
	public void report(PrintWriter out)throws IOException
	{
		if(getLogWriter()==null)return;

		out.println("<html><head><title>IndexRobot Browse Results</title></head>");
		out.println("<body bgcolor=\"#eeeeee\">");
		if(proxyInetAddress!=null)
		{
			out.println("HTTP proxy server: "+proxyInetAddress.getHostAddress()+":"+proxyPort);
			out.println("<br>");
		}
		out.println("The maximal recursive depth: "+maxDepth+"<br>");
		if(interestedIdentifiers!=null
			&& interestedIdentifiers.length>0)
		{			
			out.println("The interested identifier"+
				(interestedIdentifiers.length>1?"s are ":" is ")
				+interestedIdentifiers[0]);
			if(interestedIdentifiers.length>1)
			{
				for(int i=1;i<interestedIdentifiers.length-1;i++)
					out.println(interestedIdentifiers[i]);
				out.println(" and "+interestedIdentifiers[interestedIdentifiers.length-1]);
			}
			out.println(".<br>");
		}

		report(urlsTable,out);

		out.println("</body></html>");
	}

	//The processed result type of urls

	//Failed to visited this url for unknown reason
	public static final int UNKNOWN_LINK=0;
	/*
	 * This url is a disallowable or unwelcome URL
	 * the URL shouldn't be visited
	 */
	private static final int DISALLOWABLE_LINK=1;
	//This url return an unfamiliar message or is not visited
	private static final int UNFAMILIAR_MESSAGE=2;
	/*
	 * visited URL type
	 */
	private static final int VISITED_URL=3;
	private static final int HTTP_OK=4;
	private static final int HTTP_MOVED_PERM=5;
	private static final int HTTP_MOVED_TEMP=6;
	private static final int HTTP_BAD_REQUEST=7;
	private static final int HTTP_UNAUTHORIZED=8;
	private static final int HTTP_FORBIDDEN=9;
	private static final int HTTP_NOT_FOUND=10;
	private static final int HTTP_SERVER_ERROR=11;
	private static final int HTTP_BAD_GATEWAY=12;
	private static final int HTTP_UNAVAILABLE=13;
	private static final int MESSAGE_KIND=14;

	/*
	 * Outputs the link information.
	 * @param noticedLinks The hashtable containing all links found.
	 * @param out An output stream.
	 */
	private static final void report(Hashtable noticedLinks,PrintWriter out)
		throws IOException
	{
		LinkedList[] linksLinkedList=new LinkedList[MESSAGE_KIND];
		String[] responseMessage=new String[MESSAGE_KIND];
		for(int i=0;i<linksLinkedList.length;i++)
		{
			linksLinkedList[i] = new LinkedList();
			responseMessage[i]=null;
		}
		responseMessage[UNKNOWN_LINK]="Unknown URL";
		responseMessage[DISALLOWABLE_LINK]="Disallowable URL";
		responseMessage[UNFAMILIAR_MESSAGE]="Unfamiliar or unvisited URL";
		responseMessage[VISITED_URL]="Visited formerly URL";

		
		Object obj;
		URLInfo link;
		Enumeration links = noticedLinks.elements();
		int count=0;
		while(links.hasMoreElements())
		{
			obj=links.nextElement();
			if(obj instanceof URLInfo)
				link = (URLInfo) obj;
			else//A visited URL is not refered in this search session
				continue;
			count++;
			int type;
			switch(link.responseCode)
			{
			case HttpURLConnection.HTTP_OK:
				type=HTTP_OK;
				break;
			case HttpURLConnection.HTTP_MOVED_PERM:
				type=HTTP_MOVED_PERM;
				break;
			case HttpURLConnection.HTTP_MOVED_TEMP:
				type=HTTP_MOVED_TEMP;
				break;
			case HttpURLConnection.HTTP_BAD_REQUEST:
				type=HTTP_BAD_REQUEST;
				break;
			case HttpURLConnection.HTTP_UNAUTHORIZED:
				type=HTTP_UNAUTHORIZED;
				break;
			case HttpURLConnection.HTTP_FORBIDDEN:
				type=HTTP_FORBIDDEN;
				break;
			case HttpURLConnection.HTTP_NOT_FOUND:
				type=HTTP_NOT_FOUND;
				break;
			case HttpURLConnection.HTTP_SERVER_ERROR:
				type=HTTP_SERVER_ERROR;
				break;
			case HttpURLConnection.HTTP_BAD_GATEWAY:
				type=HTTP_BAD_GATEWAY;
				break;
			case HttpURLConnection.HTTP_UNAVAILABLE:
				type=HTTP_UNAVAILABLE;
				break;
			case UNKNOWN_LINK:
				type=UNKNOWN_LINK;
				break;
			default:
				switch(link.urlType)
				{
				case DISALLOWABLE_LINK:
					type=DISALLOWABLE_LINK;
					break;
				case VISITED_URL:
					type=VISITED_URL;
					break;
				default:
					type=UNFAMILIAR_MESSAGE;
				}
			}
			linksLinkedList[type].add(link);
			if(responseMessage[type]==null)
				responseMessage[type]=link.responseMessage;
		}

		out.println("This search session found " + count +" link"+(count>1?"s":"")+".<br>");
		for(int i=0;i<responseMessage.length;i++)
		{
			if(linksLinkedList[i].size()>0)
			{
				out.println("Echoed Response Message: "+responseMessage[i]+": " +linksLinkedList[i].size() + "<br>");
				outputLinksAndReferers(i,linksLinkedList[i],out);
			}
		}
		out.flush();
	}
	
	/*
	 * Prints a series and Links and their referers
	 *
	 * @param messageType the url mesage type.
	 * @param links An LinkedList of the links to print.
	 * @param out An output stream.
	 */
	private static final void outputLinksAndReferers(
		int messageType,LinkedList links,PrintWriter out)
		throws IOException 
	{
		URL url;
		URLInfo link;
		out.println("<ul>");
		for(int i=0;i<links.size();i++)
		{
			link = (URLInfo) links.get(i);
			out.println("<li><a href=\"" + link.url.toString()
				+"\">" + link.url.toString() +"</a>");
			if(link.keywordNumber>0)
				out.println("("+link.keywordNumber+(link.keywordNumber>1?" keywords)":" keyword)"));
			if(link.referers!=null && link.referers.size()>0)
			{
				out.println("<dd>Linked from:");
				for(int j=0;j<link.referers.size();j++)
				{
					url = (URL) link.referers.get(j);
					if(link.referers.size()==1)
						out.println("<b><a href=\"" + url.toString() + "\">" + url.toString() +
							"</a></b>");
					else
						out.println("<dd><b><a href=\"" + url.toString() + "\">" + url.toString() +
							"</a></b>");
				}
			}
			if((messageType==UNKNOWN_LINK || messageType==UNFAMILIAR_MESSAGE)
				&& link.responseMessage!=null)
			{
				out.println("<b>"+link.responseMessage+"</b>");
			}
		}
		out.println("</ul>");
	}
}

/**
 * An auxillary class which holds information related
 * to a url link.
 */
class URLInfo
{	
	URL url;
	int urlID;//Use it to quote the specific url
	int urlType;

	int keywordNumber;//The keyword number of this url

	int responseCode=-1;
	String responseMessage;

	LinkedList referers=null;//Contains all the webpages which refer this url

	int depth=Integer.MAX_VALUE;//the current recursive depth
	/* The corresponding recursive depth should also be recorded so that an extracted URL can be discarded if its recursive depth is beyond the maximal recursive depth. */

	/**
	 * Creates a URLInfo object
	 *
	 * @param url the url which will be processed
	 */
	public URLInfo(URL url)
	{
		this.url=url;
	}
	/**
	 * Creates a URLInfo object
	 *
	 * @param url the url which will be processed
	 * @param urlID the urlID of the specific url
	 * @param urlType the urlType which will be processed
	 */
	public URLInfo(URL url,int urlID,int urlType)
	{
		this.url=url;
		this.urlID=urlID;
		this.urlType=urlType;
	}
	/**
	 * Adds a referer for this URLInfo object
	 *
	 * @param referer the webpage which refer this url
	 */
	protected void addReferer(String referer)
	{
		if(referer==null)return;
		if(referer!=null)
		{
			referers=new LinkedList();
		}
		referers.add(referer);
	}
	/**
	 * Adds a referer for this URLInfo object
	 *
	 * @param referer the webpage which refer this url
	 */
	protected void setDepth(int depth)
	{
		if(this.depth>depth)this.depth=depth;
		if(this.depth<1)this.depth=1;
	}
	public String toString()
	{
		return url.toString();
	}
 }

/**
 * An auxillary class which holds keywords information related
 * to a url.
 */
class WordIDs implements java.io.Serializable
{
	transient ArrayList wordIDs;
	public WordIDs()
	{
		wordIDs=new ArrayList(5);
	}
	public WordIDs(int wordID)
	{
		wordIDs=new ArrayList(5);
		wordIDs.add(new Integer(wordID));
	}
	public WordIDs(int[] wordIDsArray)
	{
		wordIDs=new ArrayList(wordIDsArray.length);
		for(int i=0;i<wordIDsArray.length;i++)
			wordIDs.add(new Integer(wordIDsArray[i]));
	}

	private void writeObject(ObjectOutputStream out)throws IOException
	{
		out.writeInt(wordIDs.size());
		for(int i=0;i<wordIDs.size();i++)
			out.writeInt(((Integer)wordIDs.get(i)).intValue());
	}

	private void readObject(ObjectInputStream in)throws IOException,ClassNotFoundException
	{
		int size=in.readInt();
		wordIDs=new ArrayList(size);
		for(int i=0;i<size;i++)
			wordIDs.add(new Integer(in.readInt()));
	}
}